
/* 
 * simplicit -
 *     a simple implicit surface polygonizer
 *     applications should call the subroutine polygonize()
 *     translated from Mesa, a strongly typed language
 *
 * to compile a test program for ASCII output:
 *     cc -g simplicit.c -o simplicit -lm
 *
 * to compile a test program for display on an IRIS workstation:
 *     cc -DIRISGFX simplicit.c -o simplicit -lgl_s -lm
 *
 * Authored by Jules Bloomenthal, Xerox PARC.
 * Copyright (c) Xerox Corporation, 1991.  All rights reserved.
 * Permission is granted to reproduce, use and distribute this code for
 * any and all purposes, provided that this notice appears in all copies.
 *
 * From: Jules Bloomenthal <blooment@parc.xerox.com>
 * Subject: polygonizer code
 */

#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <sys/types.h>

/* #define RES     10 */ /* # converge iterations    */
#define RES     8  /* # converge iterations    */

#define L       0  /* left direction: -x, -i   */
#define R       1  /* right direction: +x, +i  */
#define B       2  /* bottom direction: -y, -j */
#define T       3  /* top direction: +y, +j    */
#define N       4  /* near direction: -z, -k   */
#define F       5  /* far direction: +z, +k    */

/* the lbn corner of cube (i, j, k), corresponding with location
 * (start.x+(i-.5)*size, start.y+(j-.5)*size, start.z+(k-.5)*size) */

#define HASHBIT     (5)
#define HASHSIZE    (size_t)(1<<(3*HASHBIT)) /* hash table size (32768) */
#define MASK        ((1<<HASHBIT)-1)
#define HASH(i,j,k) ((((((i)&MASK)<<HASHBIT)|((j)&MASK))<<HASHBIT)|((k)&MASK))

typedef struct point {             /* a three-dimensional point */
    double x, y, z;                /* its coordinates */
} POINT;

typedef struct test {              /* test function for a particular sign */
    POINT p;                       /* location of test */
    double value;                  /* function value at p */
    int ok;                        /* if value is of correct sign */
} TEST;

typedef struct vertex {            /* surface vertex */
    POINT position, normal;        /* position and surface normal */
} VERTEX;

typedef struct vertices {          /* list of vertices in polygonization */
    int count, max;                /* # vertices, max # allowed */
    VERTEX *ptr;                   /* dynamically allocated */
} VERTICES;

typedef struct corner {            /* corner of a cube */
    int i, j, k;                   /* (i, j, k) is index within lattice */
    double x, y, z, value;         /* value of function at (x,y,z) */
} CORNER;

typedef struct cube {              /* partitioning cell (cube) */
    int i, j, k;                   /* lattice location of cube */
    CORNER *lbn, *lbf, *ltn, *ltf; /* l: left,  b: bottom, n: near */
    CORNER *rbn, *rbf, *rtn, *rtf; /* r: right, t: top,    f: far */
} CUBE;

typedef struct cubes {             /* linked list of cubes acting as stack */
    CUBE cube;                     /* a single cube */
    struct cubes *next;            /* remaining elements */
} CUBES;

typedef struct centerlist {        /* list of cube locations */
    int i, j, k;                   /* cube location */
    struct centerlist *next;       /* remaining elements */
} CENTERLIST;

typedef struct cornerlist {        /* list of corners */
    int i, j, k;                   /* corner id */
    double value;                  /* corner value */
    struct cornerlist *next;       /* remaining elements */
} CORNERLIST;

typedef struct edgelist {          /* list of edges */
    int i1, j1, k1, i2, j2, k2;    /* edge corner ids */
    int vid;                       /* vertex id */
    struct edgelist *next;         /* remaining elements */
} EDGELIST;

typedef struct process {           /* parameters, function, storage */
    double (*function)();                /* implicit function */
    int (*triproc)();              /* triangle output function */
    double size, delta;            /* cube size, normal delta */
    int bounds;                    /* cube range within lattice */
    POINT start;                   /* start point on surface */
    CUBES *cubes;                  /* list of "active" cubes */
    VERTICES vertices;             /* surface vertices */
    CENTERLIST **centers;          /* cube center hash table */
    CORNERLIST **corners;          /* corner value hash table */
    EDGELIST **edges;              /* edge and vertex id hash table */
} PROCESS;

char *mycalloc();


/**** A Simple Test Program ****/

#if( 0 )
/* object: an object function */

double object (x, y, z)
double x, y, z;
{
/* star
    double x2 = x*x, y2 = y*y, z2 = z*z, z3 = z2*z;
    double x4 = x2*x2, y4 = y2*y2, z4 = z2*z2;
    return x4 + y4 + z4 - x2*y2 - x2*z2 - y2*z2 - 0.5;
*/
/* heart 
    double x2 = x*x, y2 = y*y, z2 = z*z, z3 = z2*z;
    double f = 2*x2+y2+z2-1, f3 = f*f*f;
    return f3 - 0.1*x2*z3 - y2*z3;
*/
/* wave
    return cos(x)*cos(y) + cos(y)*cos(z) + cos(x)*cos(z);
*/
}
#endif

/* torus: a torus function with major, minor radii = 0.5, 0.1 */

double torus (x, y, z)
double x, y, z;
{
    double x2 = x*x, y2 = y*y, z2 = z*z;
    double a = x2+y2+z2+(0.5*0.5)-(0.1*0.1);
    return a*a-4.0*(0.5*0.5)*(y2+z2);
}


/* sphere: an inverse square function */

double sphere (x, y, z)
double x, y, z;
{
    double rsq = x*x+y*y+z*z;
    if (rsq < 0.00001) rsq = 0.00001;
    return 1.0/rsq;
}


/* blob: a three-pole blend function */

double blob (x, y, z)
double x, y, z;
{
    return 4.3-sphere(x+1.0,y,z)-sphere(x,y+1.0,z)-sphere(x,y,z+1.0);
}


/* global variables needed by the application */

int gntris;
VERTICES gvertices;

typedef unsigned int list[3];
list gconnect[32768];

/* triangle: called by polygonize() for each triangle; write to stdout */

triangle (i1, i2, i3, vertices)
int i1, i2, i3;
VERTICES vertices;
{
    gvertices = vertices;
    gconnect[gntris][0] = i1+1;
    gconnect[gntris][1] = i3+1;
    gconnect[gntris][2] = i2+1;
    gntris++;
/*  fprintf(stdout, "3 %d %d %d\n", i1+1, i3+1, i2+1); */
    return 1;
}


/* main: call polygonize() with object function
 * write points-polygon formatted data to stdout */

#if 0 
main (argc, argv)
int argc;
char **argv;
{
    int i;
    char *err, *polygonize();
    gntris = 0;
    if ((err = polygonize(object, 0.20, 20, 0., 0., 0., triangle)) != NULL) {
        fprintf(stdout, "%s\n", err);
        exit(1);
    }
    fprintf(stdout, "%d %d %d\n", gvertices.count, gntris, 3*gntris);
    for (i = 0; i < gvertices.count; i++)
        fprintf(stdout, "%f  %f  %f\n",
            gvertices.ptr[i].position.x,
            gvertices.ptr[i].position.y,
            gvertices.ptr[i].position.z);
    for (i = 0; i < gntris; i++)
        fprintf(stdout, "3 %d %d %d\n",
            gconnect[i][0],
            gconnect[i][1],
            gconnect[i][2] );
    fprintf(stdout, "%d\n", gvertices.count);
    for (i = 0; i < gvertices.count; i++)
        fprintf(stdout, "%f  %f  %f\n",
            gvertices.ptr[i].normal.x,
            gvertices.ptr[i].normal.y,
            gvertices.ptr[i].normal.z);
}

#endif

/**** A Simple Implicit Surface Polygonizer ****/


/* polygonize: polygonize the implicit function
 *   arguments are:
 *       double function (x, y, z)
 *           double x, y, z (an arbitrary 3D point)
 *           the implicit surface function
 *           return negative for inside, positive for outside
 *       double size
 *           width of the partitioning cube
 *       int bounds
 *           max. range of cubes (+/- in three directions) from first cube
 *       double x, y, z
 *           coordinates of a "starting" point on or near the surface
 *           may be defaulted to 0, 0, 0
 *       int triproc (i1, i2, i3, vertices)
 *           int i1, i2, i3 (indices into the vertex array)
 *           VERTICES vertices the vertex array, indexed from 0)
 *           called for each triangle
 *           the triangle coordinates are (for i = i1, i2, i3):
 *               vertices.ptr[i].position.x, .y, and .z
 *           vertices are ccw when viewed from the out (positive) side
 *               in a left-handed coordinate system
 *           vertex normals point outwards
 *           return 1 to continue, 0 to abort
 *   return error or NULL
 */

char *polygonize (function, size, bounds, x, y, z, triproc)
double (*function)();
double size;
int bounds;
double x, y, z;
int (*triproc)();
{
    CORNER *setcorner();
    PROCESS p;
    CUBES *tstface();
    TEST in, out, find();

    p.function = function;
    p.triproc = triproc;
    p.size = size;
    p.bounds = bounds;
    p.delta = size/(double)(RES*RES);

    /* allocate hash tables: */
    p.centers = (CENTERLIST **) mycalloc(HASHSIZE,sizeof(CENTERLIST *));
    p.corners = (CORNERLIST **) mycalloc(HASHSIZE,sizeof(CORNERLIST *));
    p.edges =   (EDGELIST   **) mycalloc(2*HASHSIZE,sizeof(EDGELIST *));
    setcenter(p.centers, 0, 0, 0);

    srand48(1);                 /* seed random search  */
    in = find(1, &p, x, y, z);  /* find positive point */
    out = find(0, &p, x, y, z); /* find negative point */
    if (!in.ok || !out.ok)
        return "can't find starting point";

    converge(&in.p, &out.p, in.value, p.function, &p.start);

    /* push initial cube on stack: */
    p.cubes = (CUBES *) mycalloc(1, sizeof(CUBES)); /* list of 1 */
    p.cubes->cube.i = p.cubes->cube.j = p.cubes->cube.k = 0;
    p.cubes->next = NULL;

    /* 8 corners of initial cube: */
    p.cubes->cube.lbn = setcorner(&p, 0, 0, 0);
    p.cubes->cube.lbf = setcorner(&p, 0, 0, 1);
    p.cubes->cube.ltn = setcorner(&p, 0, 1, 0);
    p.cubes->cube.ltf = setcorner(&p, 0, 1, 1);
    p.cubes->cube.rbn = setcorner(&p, 1, 0, 0);
    p.cubes->cube.rbf = setcorner(&p, 1, 0, 1);
    p.cubes->cube.rtn = setcorner(&p, 1, 1, 0);
    p.cubes->cube.rtf = setcorner(&p, 1, 1, 1);

    p.vertices.count = p.vertices.max = 0; /* no vertices yet */
    p.vertices.ptr = NULL;

    while (p.cubes != NULL) { /* process active cubes till none left */
        CUBE c;
/*      CUBES *temp = p.cubes; */
        c = p.cubes->cube;
        /* polygonize the five tetrahedra in the cube: */
        if (((c.i+c.j+c.k) % 2) != 0) {          /* cube is "odd" */
            if (! tet(c.lbn, c.lbf, c.ltn, c.rbn, &p) ||
                ! tet(c.ltf, c.ltn, c.lbf, c.rtf, &p) ||
                ! tet(c.rbf, c.rbn, c.rtf, c.lbf, &p) ||
                ! tet(c.rtn, c.rtf, c.rbn, c.ltn, &p) ||
                ! tet(c.lbf, c.ltn, c.rbn, c.rtf, &p)) return "aborted";
        }
        else {                                   /* cube is "even" */
            if (! tet(c.lbf, c.rbf, c.ltf, c.lbn, &p) ||
                ! tet(c.ltn, c.ltf, c.rtn, c.lbn, &p) ||
                ! tet(c.rbn, c.rbf, c.lbn, c.rtn, &p) ||
                ! tet(c.rtf, c.rtn, c.ltf, c.rbf, &p) ||
                ! tet(c.lbn, c.ltf, c.rtn, c.rbf, &p)) return "aborted";
        };
        /* pop current cube from stack */
        p.cubes = p.cubes->next;
/*      free((char *) temp); */
        /* test six (l,r,t,b,n,f) face directions, maybe add to cube stack: */
        p.cubes = tstface(c.i-1, c.j, c.k, L, c.lbn, c.lbf, c.ltn, c.ltf, &p);
        p.cubes = tstface(c.i+1, c.j, c.k, R, c.rbn, c.rbf, c.rtn, c.rtf, &p);
        p.cubes = tstface(c.i, c.j-1, c.k, B, c.lbn, c.lbf, c.rbn, c.rbf, &p);
        p.cubes = tstface(c.i, c.j+1, c.k, T, c.ltn, c.ltf, c.rtn, c.rtf, &p);
        p.cubes = tstface(c.i, c.j, c.k-1, N, c.lbn, c.ltn, c.rbn, c.rtn, &p);
        p.cubes = tstface(c.i, c.j, c.k+1, F, c.lbf, c.ltf, c.rbf, c.rtf, &p);
    }
    return NULL;
}


/* tet: triangulate the tetrahedral simplex
 * corners presumed ordered: b, c, d appear clockwise when viewed from a
 * normally return 1, return 0 if client aborts */

int tet (a, b, c, d, p)
PROCESS *p;
CORNER *a, *b, *c, *d;
{
    int index = 0, apos, bpos, cpos, dpos;
    int e1, e2, e3, e4, e5, e6;
    if (apos = (a->value > 0.0)) index += 8;
    if (bpos = (b->value > 0.0)) index += 4;
    if (cpos = (c->value > 0.0)) index += 2;
    if (dpos = (d->value > 0.0)) index += 1;
    /* index is now 4-bit number representing one of the 16 possible cases */
    if (apos != bpos) e1 = vertid(a, b, p);
    if (apos != cpos) e2 = vertid(a, c, p);
    if (apos != dpos) e3 = vertid(a, d, p);
    if (bpos != cpos) e4 = vertid(b, c, p);
    if (bpos != dpos) e5 = vertid(b, d, p);
    if (cpos != dpos) e6 = vertid(c, d, p);
    /* 14 productive tetrahedral cases (0000 and 1111 do not yield polygons */
    switch (index) {
        case 1:  return p->triproc(e5, e6, e3, p->vertices);
        case 2:  return p->triproc(e2, e6, e4, p->vertices);
        case 3:  return p->triproc(e3, e5, e4, p->vertices) &&
                        p->triproc(e3, e4, e2, p->vertices);
        case 4:  return p->triproc(e1, e4, e5, p->vertices);
        case 5:  return p->triproc(e3, e1, e4, p->vertices) &&
                        p->triproc(e3, e4, e6, p->vertices);
        case 6:  return p->triproc(e1, e2, e6, p->vertices) &&
                        p->triproc(e1, e6, e5, p->vertices);
        case 7:  return p->triproc(e1, e2, e3, p->vertices);
        case 8:  return p->triproc(e1, e3, e2, p->vertices);
        case 9:  return p->triproc(e1, e5, e6, p->vertices) &&
                        p->triproc(e1, e6, e2, p->vertices);
        case 10: return p->triproc(e1, e3, e6, p->vertices) &&
                        p->triproc(e1, e6, e4, p->vertices);
        case 11: return p->triproc(e1, e5, e4, p->vertices);
        case 12: return p->triproc(e3, e2, e4, p->vertices) &&
                        p->triproc(e3, e4, e5, p->vertices);
        case 13: return p->triproc(e6, e2, e4, p->vertices);
        case 14: return p->triproc(e5, e3, e6, p->vertices);
    }
    return 1;
}


/* vertid: return index for vertex on edge:
 * presumed: c1->value and c2->value are of different sign
 * return saved index if any; else compute vertex and save */

int vertid (c1, c2, p)
CORNER *c1, *c2;
PROCESS *p;
{
    VERTEX v;
    POINT a, b;
    int vid = getedge(p->edges, c1->i, c1->j, c1->k, c2->i, c2->j, c2->k);
    if (vid != -1) return vid;                       /* previously computed */
    a.x = c1->x; a.y = c1->y; a.z = c1->z;
    b.x = c2->x; b.y = c2->y; b.z = c2->z;
    converge(&a, &b, c1->value, p->function, &v.position); /* position */
    vnormal(&v.position, p, &v.normal);                    /* normal */
    addtovertices(&p->vertices, v);                        /* save vertex */
    vid = p->vertices.count-1;
    setedge(p->edges, c1->i, c1->j, c1->k, c2->i, c2->j, c2->k, vid);
    return vid;
}


/* tstface: given cube at lattice (i, j, k), and four corners of face,
 * if surface crosses face, compute other four corners of adjacent cube
 * and add new cube to p's list of "active" cubes */

CUBES *tstface (i, j, k, face, c1, c2, c3, c4, p)
PROCESS *p;
int i, j, k, face;
CORNER *c1, *c2, *c3, *c4;
{
    CUBE c;
    CUBES *ret;
    CORNER *setcorner();
    int pos = c1->value > 0 ? 1 : 0;

    if ((c2->value>0) == pos && (c3->value>0) == pos && (c4->value>0) == pos)
        return p->cubes;                      /* no surface crossing */
    if (abs(i) > p->bounds || abs(j) > p->bounds || abs(k) > p->bounds)
        return p->cubes;                      /* must bound surface */
    if (setcenter(p->centers, i, j, k))
        return p->cubes;                      /* cube already visited */

    c.i = i; c.j = j; c.k = k;                /* assign new cube */
    c.lbn = c.lbf = c.ltn = c.ltf = c.rbn = c.rbf = c.rtn = c.rtf = NULL;
    switch (face) {
      case L: /* got right  (rbn,rbf,rtn,rtf), do left   (lbn,lbf,ltn,ltf) */
          c.rbn = c1; c.rbf = c2; c.rtn = c3; c.rtf = c4;
          break;
      case R: /* got left   (lbn,lbf,ltn,ltf), do right  (rbn,rbf,rtn,rtf) */
          c.lbn = c1; c.lbf = c2; c.ltn = c3; c.ltf = c4;
          break;
      case B: /* got top    (ltn,ltf,rtn,rtf), do bottom (lbn,lbf,rbn,rbf) */
          c.ltn = c1; c.ltf = c2; c.rtn = c3; c.rtf = c4;
          break;
      case T: /* got bottom (lbn,lbf,rbn,rbf), do top    (ltn,ltf,rtn,rtf) */
          c.lbn = c1; c.lbf = c2; c.rbn = c3; c.rbf = c4;
          break;
      case N: /* got far    (lbf,ltf,rbf,rtf), do near   (lbn,ltn,rbn,rtn) */
          c.lbf = c1; c.ltf = c2; c.rbf = c3; c.rtf = c4;
          break;
      case F: /* got near   (lbn,ltn,rbn,rtn), do far    (lbf,ltf,rbf,rtf) */
          c.lbn = c1; c.ltn = c2; c.rbn = c3; c.rtn = c4;
          break;
    }
    if (c.lbn == NULL) c.lbn = setcorner(p, i,   j,   k);
    if (c.lbf == NULL) c.lbf = setcorner(p, i,   j,   k+1);
    if (c.ltn == NULL) c.ltn = setcorner(p, i,   j+1, k);
    if (c.ltf == NULL) c.ltf = setcorner(p, i,   j+1, k+1);
    if (c.rbn == NULL) c.rbn = setcorner(p, i+1, j,   k);
    if (c.rbf == NULL) c.rbf = setcorner(p, i+1, j,   k+1);
    if (c.rtn == NULL) c.rtn = setcorner(p, i+1, j+1, k);
    if (c.rtf == NULL) c.rtf = setcorner(p, i+1, j+1, k+1);
    ret = (CUBES *) mycalloc(1, sizeof(CUBES));
    ret->cube = c;                             /* add to cube list */
    ret->next = p->cubes;
    return ret;
}


/* setcorner: return corner with the given lattice location
   set (and cache) its function value */

CORNER *setcorner (p, i, j, k)
int i, j, k;
PROCESS *p;
{
    /* desire speed here, so do caching here (not in separate procedure) */
    CORNER *c = (CORNER *) malloc(sizeof(CORNER));
    int index = HASH(i, j, k);
    CORNERLIST *l = p->corners[index];
    c->i = i; c->x = p->start.x+((double)i-.5)*p->size;
    c->j = j; c->y = p->start.y+((double)j-.5)*p->size;
    c->k = k; c->z = p->start.z+((double)k-.5)*p->size;
    for (; l != NULL; l = l->next)
        if (l->i == i && l->j == j && l->k == k) {
            c->value = l->value;
            return c;
        }
    l = (CORNERLIST *) malloc(sizeof(CORNERLIST));
    l->i = i; l->j = j; l->k = k;
    l->value = c->value = p->function(c->x, c->y, c->z);
    l->next = p->corners[index];
    p->corners[index] = l;
    return c;
}


/* find: search space for point with value of given sign (0: neg, 1: pos) */

TEST find (sign, p, x, y, z)
int sign;
PROCESS *p;
double x, y, z;
{
    int i;
    TEST test;
    double drand48();
    double range = p->size;
    test.ok = 1;
    for (i = 0; i < 10000; i++) {
        test.p.x = x+range*(drand48()-0.5);
        test.p.y = y+range*(drand48()-0.5);
        test.p.z = z+range*(drand48()-0.5);
        test.value = p->function(test.p.x, test.p.y, test.p.z);
        if ((sign == 0 && test.value < 0) ||
            (sign == 1 && test.value > 0)) return test;
        range = range*1.0005;        /* slowly expand search outwards */
    }
    test.ok = 0;
    return test;
}



/**** Storage ****/


/* mycalloc: return successful calloc or exit program */

char *mycalloc (nitems, nbytes)
int nitems, nbytes;
{
   char *ptr = calloc(nitems, nbytes);
   if (ptr != NULL) return ptr;
   fprintf(stderr, "can't calloc %d bytes\n", nitems*nbytes);
   exit(1);
}


/* setcenter: set (i,j,k) entry of table[]
 * return 1 if already set; otherwise, set and return 0 */

int setcenter(table, i, j, k)
CENTERLIST *table[];
int i, j, k;
{
    int index = HASH(i, j, k);
    CENTERLIST *new, *l, *q = table[index];
    for (l = q; l != NULL; l = l->next)
        if (l->i == i && l->j == j && l->k == k) return 1;
    new = (CENTERLIST *) mycalloc(1, sizeof(CENTERLIST));
    new->i = i; new->j = j; new->k = k; new->next = q;
    table[index] = new;
    return 0;
}


/* setedge: set vertex id for edge */

setedge (table, i1, j1, k1, i2, j2, k2, vid)
EDGELIST *table[];
int i1, j1, k1, i2, j2, k2, vid;
{
    unsigned int index;
    EDGELIST *new;
    if (i1>i2 || (i1==i2 && (j1>j2 || (j1==j2 && k1>k2)))) {
        int t=i1; i1=i2; i2=t; t=j1; j1=j2; j2=t; t=k1; k1=k2; k2=t;
    }
    index = HASH(i1, j1, k1) + HASH(i2, j2, k2);
    new = (EDGELIST *) mycalloc(1, sizeof(EDGELIST));
    new->i1 = i1; new->j1 = j1; new->k1 = k1;
    new->i2 = i2; new->j2 = j2; new->k2 = k2;
    new->vid = vid;
    new->next = table[index];
    table[index] = new;
}


/* getedge: return vertex id for edge; return -1 if not set */

int getedge (table, i1, j1, k1, i2, j2, k2)
EDGELIST *table[];
int i1, j1, k1, i2, j2, k2;
{
    EDGELIST *q;
    if (i1>i2 || (i1==i2 && (j1>j2 || (j1==j2 && k1>k2)))) {
        int t=i1; i1=i2; i2=t; t=j1; j1=j2; j2=t; t=k1; k1=k2; k2=t;
    };
    q = table[HASH(i1, j1, k1)+HASH(i2, j2, k2)];
    for (; q != NULL; q = q->next)
        if (q->i1 == i1 && q->j1 == j1 && q->k1 == k1 &&
            q->i2 == i2 && q->j2 == j2 && q->k2 == k2)
            return q->vid;
    return -1;
}



/**** Vertices ****/


/* addtovertices: add v to sequence of vertices */

addtovertices (vertices, v)
VERTICES *vertices;
VERTEX v;
{
    if (vertices->count == vertices->max) {
        int i;
        VERTEX *new;
        vertices->max = vertices->count == 0 ? 10 : 2*vertices->count;
        new = (VERTEX *) mycalloc((unsigned) vertices->max, sizeof(VERTEX));
        for (i = 0; i < vertices->count; i++) new[i] = vertices->ptr[i];
        if (vertices->ptr != NULL) free((char *) vertices->ptr);
        vertices->ptr = new;
    }
    vertices->ptr[vertices->count++] = v;
}


/* vnormal: compute unit length surface normal at point */

vnormal (point, p, v)
POINT *point, *v;
PROCESS *p;
{
    double f = p->function(point->x, point->y, point->z);
    v->x = p->function(point->x+p->delta, point->y, point->z);
    v->y = p->function(point->x, point->y+p->delta, point->z)-f;
    v->z = p->function(point->x, point->y, point->z+p->delta)-f;
    f = sqrt(v->x*v->x + v->y*v->y + v->z*v->z);
    if (f != 0.0) {v->x /= f; v->y /= f; v->z /= f;}
}


/* converge: from two points of differing sign, converge to zero crossing */

converge (p1, p2, v, function, p)
double v;
double (*function)();
POINT *p1, *p2, *p;
{
    int i = 0;
    POINT pos, neg;
    if (v < 0) {
        pos.x = p2->x; pos.y = p2->y; pos.z = p2->z;
        neg.x = p1->x; neg.y = p1->y; neg.z = p1->z;
    }
    else {
        pos.x = p1->x; pos.y = p1->y; pos.z = p1->z;
        neg.x = p2->x; neg.y = p2->y; neg.z = p2->z;
    }
    while (1) {
        p->x = 0.5*(pos.x + neg.x);
        p->y = 0.5*(pos.y + neg.y);
        p->z = 0.5*(pos.z + neg.z);
        if (i++ == RES) return;
        if ((function(p->x, p->y, p->z)) > 0.0)
             {pos.x = p->x; pos.y = p->y; pos.z = p->z;}
        else {neg.x = p->x; neg.y = p->y; neg.z = p->z;}
    }
}
